/**
 * 3DView
 *
 * @author KADA
 * @since 9.2.2015
 */

import { WebViewer3DAPIEXT_USE_AT_OWN_RISK } from "@esko/web-viewer-3d";
import $ from "jquery";
import _ from "underscore";
import View3DView from "./3DViewerLauncher-view";

/**
 * @param {JQueryElement} element - element where the viewer should be shown
 */
class Viewer3DLauncher extends WebViewer3DAPIEXT_USE_AT_OWN_RISK {
	constructor(element) {
		const view = new View3DView(element);
		super(view.getViewerJQElement().get(0), (arg0, arg1, arg2, arg3, arg4) => {
			this.processLoadingFeedback(arg0, arg1, arg2, arg3, arg4);
		});
		this._View3DView = view;

		// Set Transparency Level (between 1 and 10, recommended default = 7)
		// and Redraw flag (recommended default = true)
		this.SetQualityLevel(7, true);

		// Set Interaction Optimization Flag (recommended default = true)
		this.SetInteractionOptimization(true);

		//turn off reflections as it is known to cause performance issues on FF
		this.SetReflections("OFF");

		// A lot of display issues are solve by already resizing here.
		this.Resize();

		//furthermore need to register a 'window resize' callback for resizing the viewport
		//NOTE: _.throttle is used to prevent calling viewer.Resize() on every single window resize event
		$(window).on(
			"resize",
			_.throttle(() => {
				this.Resize();
			}, 40)
		);
	}

	/**
	 * Load 3D for given document version id
	 * @param {String} docVersionId - docVersion ID for which the 3D view should be shown.
	 * @param {Function} [callback] - called when 3D has loaded. First parameter indicates succes. Second the error if there is any error.
	 * @param {Object} [position] - the initial camera position. See setCameraPosition.
	 * @param {Object} [options] Options object used to customize loading of the scene
	 */
	load3DForDocVersion() {
		throw "Not Implemented: Viewer3DLauncher.prototype.load3DForDocVersion";
	}

	/**
	 * Get the current camera position if the 3D viewer is loaded or null otherwise.
	 * @returns {Object} @example
	 *                   {scale: 1.5,
	 * 					 distance: 200,
	 *                   azimuth: 46.6,
	 *                   elevation: -52,
	 *                   panx: 24.0,
	 *                   pany: -12.8,
	 *                   environment: {_x: .268, _y: -.2157, _z: .684, _w: .5498}}
	 */
	getCameraPosition() {
		var angles = this.GetCameraAngles();
		var scale = this.GetCameraScale();
		var offSet = this.GetCameraPanOffset();
		var envRot = this.GetEnvironmentRotation();
		var dist = this.GetCameraDistance();
		if (angles !== undefined && scale !== undefined && angles.azimuth !== undefined && angles.elevation !== undefined) {
			var position = {
				scale: scale,
				azimuth: angles.azimuth,
				elevation: angles.elevation,
				panx: offSet.x,
				pany: offSet.y,
				environment: envRot,
				distance: dist
			};
			return position;
		}
	}

	/**
	 * Set the current camera position of the 3D widget if the 3D viewer is loaded
	 * @param {Object} position -
	 *                 @example
	 *                    {scale: 1.5,
	 * 					  distance: 200,
	 *                    azimuth: 46.6,
	 *                    elevation: -52,
	 *                    panx: 24.0,
	 *                    pany: -12.8,
	 *                    environment: {_x: .268, _y: -.2157, _z: .684, _w: .5498}}
	 */
	setCameraPosition(position) {
		if (position) {
			const actions = [];
			if (position.scale != null) {
				actions.push(redraw => {
					this.SetCameraScale(position.scale, redraw);
				});
			}
			if (position.azimuth != null && position.elevation != null) {
				actions.push(redraw => {
					this.SetCameraAngles(position.azimuth, position.elevation, redraw);
				});
			}
			if (position.panx != null && position.pany != null) {
				actions.push(redraw => {
					this.SetCameraPanOffset(position.panx, position.pany, redraw);
				});
			}
			if (position.environment != null) {
				actions.push(redraw => {
					this.SetEnvironmentRotation(position.environment, redraw);
				});
			}
			if (position.distance > 0) {
				actions.push(redraw => {
					this.SetCameraDistance(position.distance, redraw);
				});
			}

			actions.forEach((action, i) => {
				action(i === actions.length - 1);
			});
		}
	}

	/**
	 * @deprecated Use Resize instead
	 */
	resize() {
		this.Resize();
	}

	/**
	 * Handles displaying loading progress and emitting an event when the scene is ready
	 *
	 * @param {Number} arg0
	 * @param {Number} arg1
	 * @param {Number} arg2
	 * @param {Number} arg3
	 * @param {Number} arg4
	 */
	processLoadingFeedback(arg0, arg1, arg2, arg3, arg4) {
		// When we get a progress update, postpone the automatically hiding of the progress bar.
		if (this._autoHideProgressAction) {
			clearTimeout(this._autoHideProgressAction);
		}

		// Check for error ...
		if (arg0 === 0) {
			// No error, findout progress ...
			var geometryLoadedP = arg1;
			var textureLoadedP = arg2;
			var numFilesToQuery = arg3;
			var numQueriedFiles = arg4;

			// Check if loading is done ...
			if (geometryLoadedP >= 100 && textureLoadedP >= 100 && numQueriedFiles >= numFilesToQuery) {
				// hide loading message.
				this._View3DView.hideAnyMessage();
				this._emitSceneReady();
			} else {
				// Determine percentage of files loaded.
				var fileLoadingP = (numQueriedFiles / numFilesToQuery) * 100;

				// Create global percentage number for these three criteria.
				var totalP = (geometryLoadedP + textureLoadedP + fileLoadingP) / 3;

				// Round this percentage to ...
				var rounded = parseFloat(Math.round(totalP * 10) / 10).toFixed(1);

				// Display progress...
				this._View3DView.displayLoadingMessage("(" + rounded + "%)");

				// Set that the progress update message should be hidden in three seconds.
				// This will get cancelled when this function, processLoadingFeedback, is called again within this method.
				// This way we avoid the scenario where the progress updater will keep to be shown when no new call this
				// method will be made anymore from the 3d viewer (which appears to happen on FF).
				this._autoHideProgressAction = setTimeout(
					function() {
						this._View3DView.hideAnyMessage();
					}.bind(this),
					5000
				);
			}
		} else {
			// Display progress...
			this._View3DView.displayErrorMessage("Error occured while loading 3D model. (code:" + arg0 + ")");
		}
	}

	/**
	 * Get a jQuery wrapped reference to the encapsulating div of this component.
	 *
	 * @returns {jQuery} The jQuery wrapped reference to the encapsulating div of this component.
	 */
	getJQElement() {
		return this._View3DView.getJQElement();
	}

	/**
	 * Fires all events registered with {@link onSceneReadyOnce}
	 */
	_emitSceneReady() {
		if (!this._onSceneReadyEvents) {
			return;
		}
		_.defer(() => {
			while (this._onSceneReadyEvents.length > 0) {
				this._onSceneReadyEvents.pop()();
			}
		});
	}

	/**
	 * Register a handler that will be called once when the next scene is finished loading
	 *
	 * @param {Function} handler
	 */
	onSceneReadyOnce(handler) {
		if (!this._onSceneReadyEvents) {
			this._onSceneReadyEvents = [];
		}
		this._onSceneReadyEvents.push(handler);
	}

	/**
	 * Take a screenshot of the current scene and convert it to a Blob
	 * https://stackoverflow.com/a/15754051
	 *
	 * @param {String} format
	 * @param {String} width
	 * @param {String} height
	 * @returns {Blob}
	 */
	getSnapshotAsBlob(format, width, height) {
		const base64Image = this.GetSnapshot(format, width, height);
		if (!base64Image) {
			return null;
		}
		var byteString = atob(base64Image.split(",")[1]);
		var ab = new ArrayBuffer(byteString.length);
		var ia = new Uint8Array(ab);
		for (var i = 0; i < byteString.length; i++) {
			ia[i] = byteString.charCodeAt(i);
		}
		return new Blob([ab], { type: `image/${format}` });
	}
}

/**
 * Method to automatically apply this module to an input or select element ...
 */
Viewer3DLauncher.wcrinit = function(element) {
	const launcher = new Viewer3DLauncher(element);
	launcher.load3DForDocVersion(element.dataset.docversionid);
};

/**
 * Method to automatically destroy this component once initialized for a div...
 *
 * @param div {DomElement} the DomElement to destroy again
 */
Viewer3DLauncher.wcrdestroy = function() {};

export default Viewer3DLauncher;
