/**
* Extension of the WebViewer3DAPI
*
* @module WebViewer3D
* @author DOGO, PHST
* @copyright 2015 Esko Stonecube Ltd. All rights reserved.
* @version 1.0.x
*/

define([
	'./WebViewer3DAPI',
    "./modules/LogMessage"
    ], function (
    	WebViewer3DAPI,
        LogMessage
        ) {

    'use strict';

	/**
     * The WebViewer3DAPIEXT object.
     *
     *  @function WebViewer3DAPIEXT
     *  @alias module:WebViewer3D
     *  @class WebViewer3DAPIEXT
     *  @param {HTMLObject} context - a reference to the DOM element in which the viewer should be inserted
     */
    const WebViewer3DAPIEXT = function(args)
    {
        WebViewer3DAPI.apply(this,arguments);
        LogMessage.Log("Creating WebViewer3DAPIEXT Object");
    };
    WebViewer3DAPIEXT.prototype = Object.create(WebViewer3DAPI.prototype);
    WebViewer3DAPIEXT.prototype.constructor = WebViewer3DAPIEXT;

    /**
     * Reload all textures
     *
     * @function ReLoad all textures (only material textures) that have been loaded by the json file.
     * @param {function} __callback - callback funtion used when rendering has been updated
     * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
     */
    WebViewer3DAPIEXT.prototype.ReloadAllTextures = function(__callback, __redraw)
    {
        LogMessage.Log("API Reload all textures");

        this._controller.ReloadAllTextures(__callback, __redraw);
    };

    /**
     * Reload a texture
     *
     * @function ReloadTexture
     * @param {String} __textureURL - URL identifying the initial texture relative path from json to texture
     * @param {String} __newURL - URL identifying the new texture relative path from json to texture
     * @param {function} __callback - callback funtion used when rendering has been updated
     * @param {Boolean} [__keep_old_path = true] - indicates if the old VS new path has to be used now on to reference the texture
     * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
     */
    WebViewer3DAPIEXT.prototype.ReloadTexture = function(__textureURL, __newURL, __callback, __keep_old_path, __redraw)
    {
        LogMessage.Log("API Reload texture");

        this._controller.ReloadTexture(__textureURL, __newURL, __callback, __keep_old_path, __redraw);
    };

    /**
     * Set the backdrop texture
     *
     * @function SetBackdropTexture
     * @param {String} __textureURL - URL identifying the texture URL to be used
     * @param {function} __successCallback - callback funtion used when rendering has been updated
     * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
     */
    WebViewer3DAPIEXT.prototype.SetBackdropTexture = function(__textureURL, __successCallback, __redraw)
    {
        LogMessage.Log("API SetBackdrop texture");

        this._controller.SetBackdropTexture(__textureURL, __successCallback, __redraw);
    };

    /**
     * Get the queried data info of an object using its URL
     *
     * @function Get
     * @param {string} __nodeURL -  the type of information we want to retrieve from object
     * @param {string} __datatype -  the type of information we want to retrieve from object
     * to choose among the following list:
     * - NODEOBJECTVIS returns (Boolean)
     * @returns {object|boolean|string} depending on the type of info
     */
    WebViewer3DAPIEXT.prototype.Get = function(__nodeURL, __datatype)
    {
        return this._controller.Get(__nodeURL, __datatype);
    };

    /**
     * Prints a memory report
     * @function LogMemoryReport
     */
    WebViewer3DAPIEXT.prototype.LogMemoryReport = function(__object)
    {
        return this._controller.LogMemoryReport();
    };

    /**
     * Print list of opaque scene vertices in world coordinates
     */
    WebViewer3DAPIEXT.prototype.PrintOpaqueSceneVertices = function()
    {
        return this._controller.PrintOpaqueSceneVertices();
    };

    /**
     * Calculates the camera distance corresponding with a legacy scale value
     * @function ConvertLegacyScaleToDistance
     * @param {float} __scale the legacy scale value
     */
    WebViewer3DAPIEXT.prototype.ConvertLegacyScaleToDistance = function(__scale)
    {
        return this._controller.ConvertLegacyScaleToDistance(__scale);
    };

    /**
     * Load a scene from a local JSON file containing the scene description
     *
     * @function LoadSceneFromLocalFile
     * @param {File} __file - A File API file object
     * @param {Number} __coplanarityEpsilon [0.02] - if > 0 used to tune tolerance for
     * smooth surfaces so that these are considered as one single highlighted surface, i.e., only
     * their silhouette will appear.
     * @param {function} __loadSceneProgressCallback [undefined] - See the API constructor (WebViewer3DAPI) for an explanation.
     * @param {function} __clientUIAddElementCallback [undefined] - default callback function
     *            giving the current node path (arg[1]) and a boolean indicating
     *            if it is a leaf (arg[2])
     * @param {Number} __boundaryhardedges [true] - if true, builds the hard edges highlights at surface boundaries
     * @param {String} _defaultURL [null] - this allows to set the default URL path used for external file referenced by the local file
     */
    WebViewer3DAPIEXT.prototype.LoadSceneFromLocalFile = function (__file, __coplanarityEpsilon, __loadSceneProgressCallback, __clientUIAddElementCallback, __boundaryhardedges, _defaultURL)
    {
        LogMessage.Log("API Load Scene from local file");

        const reader = new FileReader();

        reader.onload = function (e)
        {
            this.LoadSceneFromObject(JSON.parse(reader.result), __coplanarityEpsilon, __loadSceneProgressCallback, __clientUIAddElementCallback, __boundaryhardedges, _defaultURL);
        }.bind(this);

        reader.readAsText(__file);
    };

    /**
     * Test function
     */
    WebViewer3DAPIEXT.prototype.APIEXTTestFunction = function(__enableAnimations)
    {
        // If needed to call super implem.
        // WebViewer3DAPI.prototype.toImplement.call(this, "myparam");
    };

    return WebViewer3DAPIEXT;

});
