/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define(["backbone", "../libs/stacktrace"], function(bb, Stacktrace) {
	"use strict";

	const GetStackTrace = function() {
		return Stacktrace.get();
	};

	/**
	 * This is a helper object to avoid code duplication in the LogMessage class itself.
	 * @param {Function} __defaultLogFunction: A function that can log a string message to somewhere (e.g. console.log, console.error,...).
	 * @param {String} __type: One of "message", "warning", "error".
	 */
	const Logger = bb.Model.extend({
		initialize: function(__defaultLogFunction, __type) {
			this.set("_clientMessageCallback", null);
			this.set("_defaultLogFunction", __defaultLogFunction); // Used when no client callback is set.
			this.set("_type", __type);
		},

		SetCallbackFunction: function(__clientMessageCallback) {
			this.set("_clientMessageCallback", __clientMessageCallback);
		},

		Log: function(__message) {
			if (this.get("_clientMessageCallback") && typeof this.get("_clientMessageCallback") === "function") {
				this.get("_clientMessageCallback")({
					string: __message,
					type: this.get("_type"),
					stackTraceFunction: GetStackTrace
				});
			} else {
				// By default we do not output the stack trace or the type. The user can still set his or her own callback using SetCallbackFunction
				this.get("_defaultLogFunction")(__message);
			}
		}
	});

	const DefaultConsoleLog = function(__message) {
		console.log(__message);
	};

	const DefaultConsoleWarning = function(__message) {
		console.warn(__message);
	};

	const DefaultConsoleError = function(__message) {
		console.error(__message);
	};

	/**
	 * @alias module:WebViewer3D
	 * @class LogMessage
	 * Warning : this class is a singleton, any modification to one of its component will change for each reference to it.
	 */
	const LogMessage = bb.Model.extend({
		initialize: function() {
			this.set("_message_logger", new Logger(DefaultConsoleLog, "message"));
			this.set("_warning_logger", new Logger(DefaultConsoleWarning, "warning"));
			this.set("_error_logger", new Logger(DefaultConsoleError, "error"));
		},

		/**
		 * Sets the callbacks for the different logging types (message, warning and error).
		 * With these callbacks the client can customize how messages are displayed.
		 * The callback is a function with one single object as argument, this object has the following keys (string, type, stackTraceFunction).
		 *     1. string is a string that contains the log message.
		 *     2. type can be either "message, "warning, or "error"
		 *     3. stackTraceFunction is a function that can be called to get the relevant StackTrace object,
		 *              that contains some information about for instance the file and line number where the message was logged.
		 *
		 * @param __clientLogMessageCallbacks: the client callback.
		 */
		SetCallbackFunctions: function(__clientLogMessageCallback) {
			if (__clientLogMessageCallback && typeof __clientLogMessageCallback === "function") {
				console.log("LogMessage: Callback function for logs is set (the following messages will use this callback function).");
				this.get("_message_logger").SetCallbackFunction(__clientLogMessageCallback);
				this.get("_warning_logger").SetCallbackFunction(__clientLogMessageCallback);
				this.get("_error_logger").SetCallbackFunction(__clientLogMessageCallback);
			} else {
				console.warn(
					"LogMessage: The given client callback to handle log messages is not a function. A default logging handler will be used instead."
				);
			}
		},

		Log: function(__message) {
			this.get("_message_logger").Log(__message);
		},

		Warning: function(__message) {
			this.get("_warning_logger").Log(__message);
		},

		Error: function(__message) {
			this.get("_error_logger").Log(__message);
		}
	});

	return new LogMessage();
});
