/**
 * @author DOGO
 * @module WebViewer3D
 */

define([
    "backbone",
    "./LogMessage"
], function (
    bb,
    LogMessage
)
{

    /**
     * Return an unique identifier every time it is called
     * @returns {string}
     */
    const uuid = function()
    {
        let deltaT = new Date().getTime();
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c)
        {
            const r = (deltaT + Math.random()*16)%16 | 0;
            deltaT = Math.floor(deltaT/16);
            return (c ==='x' ? r :(r&0x3|0x8)).toString(16);
        });
    };

    /**
     * @class MaterialFactory
     * This class is a singleton containing helper functions for Material and Texture assets definition.
     */
    const MaterialFactory = bb.Model.extend({

        /**
         * This is a convenience function that tries to determine the type of the texture given its name
         * this texture type can then be used in function SetTextureURL.
         * Note that this will work with texture names that come out of PMATE only or printable parts textures
         * that were converted using WV3DSConversion component.
         * Therefore, this does not work reliably on texture names that are generated by any other process
         * or textures that belong to non printable parts for which the user is supposed to know what type these are.
         * The fallback type will be "diffuse".
         * @param {String} __textureURL
         * @returns {String} containing the type "diffuse", "specular", "normal_map", "internal_specular", "diffraction_grating", or "diffraction_grating_filter"
         */
        GetTextureType: function (__textureURL)
        {
            if (typeof __textureURL === "string")
            {
                let filename = __textureURL.substring(__textureURL.lastIndexOf('/') + 1);//extract file name
                filename = filename.split('.').slice(0, -1).join('.');//remove extension
                if (filename.endsWith("_calcRgba")) return "diffuse";
                if (filename.endsWith("_calcSpecRgbH")) return "specular";
                if (filename.endsWith("_calcNormalMap")) return "normal_map";
                if (filename.endsWith("_calcSpecInternalRgbH")) return "internal_specular";
                if (filename.endsWith("_calcDiffGrating")) return "diffraction_grating";
                if (filename.endsWith("_calcDiffGratingfilter")) return "diffraction_grating_filter";
            }
            return "diffuse";
        },

        /**
         * Return the texture key type according from the given type
         * @param __textureType
         * @returns {string}
         * @constructor
         */
        GetTextureTypeKey: function(__textureType)
        {
            switch (__textureType)
            {
                case "diffuse": return "diffuse_tex";
                case "specular": return "specular_tex";
                case "internal_specular": return "int_specular_tex";
                case "normal_map": return "normal_map_tex";
                case "diffraction_grating_filter": return "diff_grating_filter_tex";
                case "diffraction_grating": return "diff_grating_tex";
                default: return null;
            }
        },

        /**
         * Defines a blank default material asset (white with no textures yet)
         *
         * @returns {Object} an asset of type "Material"
         */
        DefineDefaultMaterial: function()
        {
            return {
                AssetType: "Material",
                LinkIdentifer: "",//unused
                Properties:
                    {
                        reflectance_model: "Visualizer",//unused
                        shader: null,
                        double_sided: true,
                        front:
                            {
                                diffuse: [0.8, 0.8, 0.8],
                                specular: [0.7, 0.7, 0.7],
                                ambient: [1, 1, 1],
                                emission: [0, 0, 0],
                                shininess: 0.6,

                                grating_spacing_scale: 1.6,
                                default_diff_haze: 0.8,
                                default_diff_spacing: 0.4,
                                double_grating: false
                            },
                        opacity: 1,
                        absorbency: 0,//unused
                        metallic: 0//unused
                    }
            };
        },

        /**
         * Defines a texture asset from its definition
         *
         * @param __textureDefinition the texture definition (such as described in the API of DefineMaterialList)
         * @returns {Object} an asset of type "Texture"
         */
        DefineTexture: function (__textureDefinition)
        {
            const textureAsset =
                {
                    AssetType: "Texture",
                    Properties:
                        {
                            filepath: __textureDefinition.url,
                            transparent: __textureDefinition.transparent || false,
                            remapped_alpha_channel: __textureDefinition.remapped_alpha_channel || false,
                            swrap: __textureDefinition.swrap || "GL_REPEAT",
                            twrap: __textureDefinition.twrap || "GL_REPEAT"
                        }
                };

            if (__textureDefinition.baseURL)
            {
                textureAsset.Properties["baseURL"] = __textureDefinition.baseURL;
            }
            if (__textureDefinition.original_width)
            {
                textureAsset.Properties["original_width"] = __textureDefinition.original_width;
            }
            if (__textureDefinition.original_height)
            {
                textureAsset.Properties["original_height"] = __textureDefinition.original_height;
            }

            return textureAsset;
        },

        /**
         * Defines a set of textures from a material name and the asset pool containing that material
         * and insert these into the given asset pool. Also returns the modified asset pool.
         *
         * @precondition __assetPool should contain the __materialName material
         * @param __materialName the material name to which the textures are added to
         * @param __texturesDefinitions the textures definitions (see API of DefineMaterialList)
         * @param __assetPool the target pool which will contain all texture assets
         * @returns {*} the asset pool
         */
        DefineTextures: function (__materialName, __texturesDefinitions, __assetPool)
        {
            for (let i = 0; i < __texturesDefinitions.length; ++i)
            {
                const texture = __texturesDefinitions[i];
                const textureURL = texture.url;

                texture.type = texture.type || this.GetTextureType(textureURL);//type may be missing from the definition
                const textureTypeKey = this.GetTextureTypeKey(texture.type);
                if(textureTypeKey)
                {
                    const textureKey = __materialName + "_" + textureTypeKey + "_" + uuid();//Generate a unique name

                    __assetPool[textureKey] = this.DefineTexture(texture);
                    __assetPool[__materialName].Properties.front[textureTypeKey] = textureKey;
                }
                else
                {
                    LogMessage.Error("Texture type "+texture.type+" is not supported");
                }
            }
            return __assetPool;
        },

        /**
         * Defines a textured material from a name and its textures definitions
         * and insert these into the given asset pool. Also returns the modified asset pool.
         *
         * @param __materialName the material name to which the textures are added to
         * @param __texturesDefinitions the textures definitions (see API of DefineMaterialList)
         * @param __assetPool the target pool which will contain all texture assets
         * @returns {*} the asset pool
         */
        DefineTexturedMaterial: function (__materialName, __texturesDefinitions, __assetPool)
        {
            let assetPool = __assetPool || {};

            assetPool[__materialName] = this.DefineDefaultMaterial();
            assetPool = this.DefineTextures(__materialName, __texturesDefinitions, assetPool);

            return assetPool;
        },

        /**
         * Defines a set of textured materials from their names and their textures definitions
         * and return these as an asset pool
         *
         * @param __materialsDefinitions (see API of DefineMaterialList)
         * @returns the asset pool containing the materials and texture assets
         */
        DefineTexturedMaterials: function (__materialsDefinitions)
        {
            let assetPool = {};

            for(let i = 0; i < __materialsDefinitions.length; ++i)
            {
                const materialDefinition = __materialsDefinitions[i];
                if(materialDefinition.hasOwnProperty("name") &&
                    materialDefinition.hasOwnProperty("textures") &&
                    Array.isArray(materialDefinition.textures))
                {
                    assetPool = this.DefineTexturedMaterial(materialDefinition.name, materialDefinition.textures, assetPool);
                }
            }

            return assetPool;
        }
    });

    return new MaterialFactory;
});
