/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2018 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "backbone",
    "./MaterialFactory",
    "./LogMessage",
    "./URLUtils",
], function (
    bb,
    MaterialFactory,
    LogMessage,
    URLUtils
)
{
    "use strict";

    const MaterialList = bb.Model.extend({

        /**
         * @alias module:WebViewer3D
         * @class MaterialList
         * Create a MaterialList from any material.json Object.
         * The assets present in this pool will either have their AssetType flag "Material" or "Texture"
         *
         * @param {Object} __attributes object containing following properties.
         * "_JSONAssetPool": An object containing materials assets (and corresponding textures assets). This has the same structure as the material.json file (not a MaterialList).
         * "_texturesBaseURL": Absolute URL pointing to the "textures" folder (i.e. the folder that contains the textures present in the _JSONAssetPool).
         * "frontMaterialName" if provided, replaces the "front" material key name with the provided name. This should come from the _eg_metadata.json.
         * "backMaterialName" if provided, replaces the "back" material key name with the provided name. This should come from the _eg_metadata.json.
         *
         * @constructor {MaterialList} an object containing the materials (and corresponding textures) that are used by the viewer. __JSONAssetPool attribute
         *                              has the same structure as any material.json file, except that this object has the type MaterialList augmented with
         *                              utility functions such as SetTextureURL.
         *
         *         This object can be used by functions: LoadScene, LoadSceneFromObject, and UpdateMaterials in WebViewer3DAPI
         *
         */
        initialize: function (__attributes)
        {
            for (let i = 0, keys = Object.keys(__attributes._JSONAssetPool), length = keys.length; i < length; ++i)
            {
                // For textures, we need to set the base URL to be able to find textures back that are referenced by a relative path.
                // Also check if this class constructor is used properly.
                const jsonAsset = __attributes._JSONAssetPool[keys[i]];
                if (jsonAsset.AssetType === "Texture")
                {
                    if (URLUtils.IsRelativeURL(jsonAsset["Properties"]["filepath"]))
                    {
                        if (!__attributes._texturesBaseURL)
                            LogMessage.Warning("When a texture asset has a relative file path, then the textures folder base URL should be filled in in the MaterialList constructor.");
                        else
                            jsonAsset["Properties"]["baseURL"] = __attributes._texturesBaseURL;
                    }
                }
            }

            if (__attributes.frontMaterialName)
            {
                if (__attributes._JSONAssetPool.hasOwnProperty(__attributes.frontMaterialName))
                    LogMessage.Error("Asset " + __attributes.front + " is already present in the asset pool. \"front\" cannot be renamed to that name.");
                else if (__attributes._JSONAssetPool.hasOwnProperty("front"))
                {
                    __attributes._JSONAssetPool[__attributes.frontMaterialName] = __attributes._JSONAssetPool.front;
                    delete __attributes._JSONAssetPool.front;
                }
            }

            if (__attributes.backMaterialName)
            {
                if (__attributes._JSONAssetPool.hasOwnProperty(__attributes.backMaterialName))
                    LogMessage.Error("Asset " + __attributes.back + " is already present in the asset pool. \"back\" cannot be renamed to that name.");
                else if (__attributes._JSONAssetPool.hasOwnProperty("back"))
                {
                    __attributes._JSONAssetPool[__attributes.backMaterialName] = __attributes._JSONAssetPool.back;
                    delete __attributes._JSONAssetPool.back;
                }
            }
        },

        /**
         * Returns an array all materials names available in this list
         * @returns {Array} array containing the different material names
         */
        GetMaterialNames: function ()
        {
            const material_names = [];
            for (let i = 0, keys = Object.keys(this.get("_JSONAssetPool")), l = keys.length; i < l; ++i)
            {
                const materialName = keys[i];
                if (this.get("_JSONAssetPool")[materialName].AssetType === "Material")
                {
                    material_names.push(materialName);
                }
            }
            return material_names;
        },

        /**
         * This function makes sure to clean-up the structure if there are "dangling" textures
         *
         */
        CleanupDanglingTextures: function ()
        {
            //Cleanup the JSONAssetpool to remove dangling textures
            const cleanedMaterials = this.GetMaterials(this.GetMaterialNames());
            this.set("_JSONAssetPool", cleanedMaterials.get("_JSONAssetPool"));
        },

        /**
         * Returns an array of all possible texture types available in the given material
         * @param {String} __materialName
         * @returns {Array}
         */
        GetTextureTypes: function (__materialName)
        {
            const possible_types = ["diffuse", "specular", "normal_map", "internal_specular", "diffraction_grating", "diffraction_grating_filter"];
            const result_types = [];
            const material = this.get("_JSONAssetPool")[__materialName];

            if (!material)
            {
                LogMessage.Error("Material " + __materialName + " not found in list");
                return result_types;
            }
            if (material.AssetType !== "Material")
            {
                LogMessage.Error("Asset " + __materialName + " is not a material");
                return result_types;
            }

            for (let i = 0; i < possible_types.length; ++i)
            {
                const type = possible_types[i];
                if (material.Properties.front.hasOwnProperty(MaterialFactory.GetTextureTypeKey(type)))
                {
                    result_types.push(type);
                }
            }
            return result_types;
        },

        /**
         * Updates the properties of the textures of a given material(__materialName)
         * The texture type type must be either "diffuse", "specular", "normal_map", "internal_specular", "diffraction_grating", or "diffraction_grating_filter"
         * If the type is omitted, then the viewer tries to determine the type of texture using its filename (from URL, this is in some cases unreliable, please
         * see GetTextureType above for details about it).
         *
         * If the texture of the given type does not exist for this material, the texture asset is then created and added to this material list.
         * User can then use the main WebViewerAPI::UpdateMaterials function to reload that material using the new texture.
         *
         * Note that this call does not remove other texture types that this material uses.
         *
         * @param __materialName
         * @param {Array} __textures array of objects, each containing the following keys:
         *                  "url": mandatory string containing the absolute URL of the texture or the relative path if baseURL is filled in
         *                  "type": recommended string (see documentation above)
         *                  "baseURL": mandatory/unused string containing the absolute URL of the texture folder (mandatory if url is a relative path, unused if url is absolute)
         *                  "transparent": optional boolean indicating if the alpha channel of the texture contains at least one value < 255
         *                  "remapped_alpha_channel": optional boolean indicating if the alpha channel is mapped from [127-255] instead of [0-255] (fix for IE)
         *                  "swrap": optional string indicating wrapping mode for first texture coordinate. Can be "GL_REPEAT" (default), "GL_CLAMP", "GL_CLAMP_TO_BORDER", or "GL_CLAMP_TO_EDGE"
         *                  "twrap": optional string indicating wrapping mode for second texture coordinate. Can be "GL_REPEAT" (default), "GL_CLAMP", "GL_CLAMP_TO_BORDER", or "GL_CLAMP_TO_EDGE"
         *                  "original_width": optional original texture width before padding ( if it has been padded to a power of two for example)
         *                  "original_height": optional original texture height before padding ( if it has been padded to a power of two for example)
         */
        SetTextures: function (__materialName, __textures)
        {
            let material = this.get("_JSONAssetPool")[__materialName];
            if (!material)
            {
                LogMessage.Warning("Material "+ __materialName + "not found. The viewer will create one for you.");
                material = this.get("_JSONAssetPool")[__materialName] = MaterialFactory.DefineDefaultMaterial();
            }

            for (let i = 0; i < __textures.length; ++i)
            {
                const textureURL = __textures[i].url;
                let textureType = __textures[i].type || MaterialFactory.GetTextureType(textureURL);

                const textureTypeKey = MaterialFactory.GetTextureTypeKey(textureType);
                let textureKey = material.Properties.front[textureTypeKey];
                if (!textureKey)
                {
                    LogMessage.Warning("Texture of type " + textureType + " not found for material " + __materialName + ". The viewer will create one for you.");
                    textureKey = __materialName + "_" + textureTypeKey + "_" + uuid();//Generate a unique name
                    material.Properties.front[textureTypeKey] = textureKey;
                }
                else
                {
                    const textureAsset = this.get("_JSONAssetPool")[textureKey];
                    __textures[i]["filepath"] = textureURL;
                    __textures[i]["transparent"] = __textures[i].transparent || textureAsset.Properties.transparent;
                    __textures[i]["remapped_alpha_channel"] = __textures[i].remapped_alpha_channel || textureAsset.Properties.remapped_alpha_channel;
                    __textures[i]["swrap"] = __textures[i].swrap || textureAsset.Properties.swrap;
                    __textures[i]["twrap"] = __textures[i].twrap || textureAsset.Properties.twrap;
                }
                this.get("_JSONAssetPool")[textureKey] = MaterialFactory.DefineTexture(__textures[i]);
            }
        }

    });

    //We define GetMaterials function afterwards since MaterialList is not yet defined and we cannot not call "new MaterialList" otherwise

    /**
     * This convenience function allows the user to retrieve only the listed materials that he/she does want to keep
     * from this materiallist. For example, someone can get the full material list using WebViewerAPI.js::GetMaterialList
     * to retrieve this material list, but then updates one material using SetTextureURL("MyMaterial"); and then he/she
     * wants to call WebViewerAPI::UpdateMaterials using that list.
     * This function takes care of creating a new list with only the sub set of material that are given in __extractionList
     * so that WebViewerAPI::UpdateMaterials can take less time to update all the materials.
     * @param {Array} __extractionList an array of strings, each referring a material in this material list.
     * @returns {MaterialList} a new material list, composed with only the subset of specified materials
     */
    MaterialList.prototype.GetMaterials = function (__extractionList)
    {
        const newMaterialPool = {};
        for (let i = 0; i < __extractionList.length; ++i)
        {
            const materialName = __extractionList[i];
            const material = this.get("_JSONAssetPool")[materialName];
            if (!material)
            {
                LogMessage.Error("Material " + materialName + " not found in list");
                continue;
            }
            if (material.AssetType !== "Material")
            {
                LogMessage.Error("Asset " + materialName + " is not a material");
                continue;
            }

            newMaterialPool[materialName] = $.extend(true, {}, material);//deep copy of material asset
            const types = this.GetTextureTypes(materialName);
            for (let j = 0; j < types.length; ++j)
            {
                const textureKey = material.Properties.front[MaterialFactory.GetTextureTypeKey(types[j])];
                const texture = this.get("_JSONAssetPool")[textureKey];

                if (!texture)
                {
                    LogMessage.Error("Texture " + textureKey + " not found in list");
                    continue;
                }
                if (texture.AssetType !== "Texture")
                {
                    LogMessage.Error("Asset " + textureKey + " is not a texture");
                    continue;
                }

                newMaterialPool[textureKey] = $.extend(true, {}, texture);//deep copy of texture asset
            }
        }
        return new MaterialList({_JSONAssetPool:newMaterialPool});
    };

    return MaterialList;

});
