/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "./AssetManager",
    "./LogMessage",
    "./CallbackWrapper"
], function (
        _,
        bb,
        WV3DAssetManager,
        LogMessage,
        CallbackWrapper
        )
{

    'use strict';

    /**
     * @alias module:WebViewer3D
     * @class WV3DViewerModel
     */
    return bb.Model.extend({

        /**
         * @param {Function} __loadSceneProgressCallback:
         *          Called when progress updates are available during the scene loading.
         *          The signature of the callback is expected to be (__error, __geometry_load_percent, __texture_load_percent, __queried_assets_nb, __resolved_assets_nb).
         */
        initialize: function (__loadSceneProgressCallback, __urlPreprocessCallback)
        {
            LogMessage.Log("Creating WV3DViewerModel Object");

            this.set('_assetmanager', new WV3DAssetManager(this.LoadSceneJSON.bind(this), this.ReportLoadError.bind(this), __urlPreprocessCallback));
            this.set('SceneId', null);

            if (typeof __loadSceneProgressCallback === "function")
                this.set('_loadSceneProgressCallback', __loadSceneProgressCallback);
            else
                this.set('_loadSceneProgressCallback', null);
        },

        /**
         * Query a scene from a url and gives LoadSceneJSON (next function) as a callback to load
         * this scene when the resource is available
         */
        LoadSceneURL: function (__sceneURL)
        {
            LogMessage.Log("Model LoadSceneURL function");

            this.get('_assetmanager').LoadURL(__sceneURL);
        },

        /**
         * Load a scene from a json file and trigger an event to tell the viewer that the model
         * has been updated. This function will be called once for each nested json file by the
         * httprequestmanager.
         *
         * Also call the client's callback function to indicate state of scene loading. Argument
         * indicates loading state, 0 for success and successive values to indicate errors.
         * Description of these values is provided in 3DAPI JS.
         *
         * @param __resource
         * @param __sceneURL will replace the AssetID key to reference to that scene
         */
        LoadSceneJSON: function (__resource, __sceneURL)
        {
            LogMessage.Log("Model LoadSceneJSON function");

            //IMPROVEME: change scene id handling
            // The first time we come in this function, we need to set the scene id and insert the replacement materials to the asset pool.
            const scene_id = this.get('_assetmanager').LoadAsset(__resource, __sceneURL);
            if (this.get('SceneId') == null)
            {
                this.set('SceneId', scene_id);
                this.get('_assetmanager').InsertMaterialsInAssetPool(this.get('SceneId'));
            }

            CallbackWrapper(this.get('_loadSceneProgressCallback'))(0, 0, 0, this.get('_assetmanager').get('_queriedassetnumber'), this.get('_assetmanager').get('_resolvedassetnumber'));

            this.trigger('added_scene_resource');
        },

        /**
         * See the UpdateMaterials function in WebViewer3DAPI.js for the documentation.
         */
        UpdateMaterials : function (__materials, __forceRebuild)
        {
            LogMessage.Log("Model UpdateMaterials function");

            // check if materials are compatible with updating (no changes in holo/transparency)
            if ( !__forceRebuild && this.get("_assetmanager").CheckReplacementMaterialsForCompatibility(this.get('SceneId') , __materials))
            {
                this.get("_assetmanager").SetReplacementMaterials(__materials);
                this.get("_assetmanager").InsertMaterialsInAssetPool(this.get("SceneId"));
                this.trigger("update_materials");
            }
            else
            {
                LogMessage.Log("Incompatible materials when updating, triggering full reload");
                this.get("_assetmanager").SetReplacementMaterials(__materials);
                this.get("_assetmanager").InsertMaterialsInAssetPool(this.get("SceneId"));
                this.trigger("rebuild_scene");
            }
        },

        /**
         * See the CreateMaterialList function in WebViewer3DAPI.js for the documentation.
         */
        CreateMaterialList : function (__materials, __texturesBaseURL, __frontMaterialName, __backMaterialName)
        {
            LogMessage.Log("Model CreateMaterialList function");

            return this.get("_assetmanager").CreateMaterialList(__materials, __texturesBaseURL, __frontMaterialName, __backMaterialName);
        },

        /**
         * See the DefineMaterials function in WebViewer3DAPI.js for the documentation.
         */
        DefineMaterialList : function (__materialsDefinitions)
        {
            LogMessage.Log("Model DefineMaterials function");

            return this.get("_assetmanager").DefineMaterialList(__materialsDefinitions);
        },

        /**
         * Error report to API user when an error occured
         */
        ReportLoadError: function (__resourceURL)
        {
            LogMessage.Log("Model Asset manager could not load queried resource: " + __resourceURL);

            CallbackWrapper(this.get('_loadSceneProgressCallback'))(2, 0, 0, this.get('_assetmanager').get('_queriedassetnumber'), this.get('_assetmanager').get('_resolvedassetnumber'));
        },

        /**
         * Delete the current scene
         */
        ResetAssetManager: function ()
        {
            LogMessage.Log("Reset Model function");

            this.get('_assetmanager').Clear();
            this.set('SceneId', null);
        },

        SetDefaultResourceRelativePath: function (__path)
        {
            this.get('_assetmanager').SetDefaultResourceRelativePath(__path);
        }
    });
});
