import ViewportPopoverModel from "@esko/web-viewer/viewport/model/ViewportPopoverModel";
import tpl from "@esko/web-viewer/svg/TextSelectionPopoverTemplate.tpl.html";
import Template from "@esko/template";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
import $ from "jquery";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
/**
 * PoveroverModel for text selection tool
 *
 * @exports TextSelectionPopoverModel
 * @memberof SvgModule
 * @class TextSelectionPopoverModel
 * @augments Viewport.Model.ViewportPopoverModel
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {String} name
 */
var TextSelectionPopoverModel = function(f, name, actions) {
	ViewportPopoverModel.call(this, f, name);

	this._content = $("<div/>").addClass("TextSelectionPopover");
	this.textContent = "";
	this._popoverLocation = null;
	this._worldTextFrames = null;
	this._boundingWorldFrame = null;

	this._options = {};

	// initialize template
	this.t = new Template(f.i18n, tpl);
	this.t.setContext(this._content);

	// Add default action: Copy selection
	this.t.bind(
		"mousedown",
		".copyText",
		_.bind(function(source, event) {
			const el = document.createElement("textarea");
			el.value = this._textContent;
			document.body.appendChild(el);
			el.select();
			document.execCommand("copy");
			document.body.removeChild(el);

			this._emit("popupClosed");
		}, this)
	);

	// Add registered listeners to the template
	_.each(actions, (listener, name) => {
		const className = "." + name;
		this.t.bind("mousedown", className, (source, event) => {
			listener(this._textContent, this._boundingWorldFrame, this._worldTextFrames);
			this._emit("popupClosed");
		});
	});

	// load language if needed
	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.t.render();
		}, this)
	);

	this.getFrame = function() {
		if (this._popoverLocation) {
			return this._popoverLocation;
		} else {
			return new ShapeUtils.Rect(new ShapeUtils.Point(0, 0), new ShapeUtils.Size(1, 1));
		}
	};

	this.getContent = function() {
		return this._content;
	};

	this.getPreferredPosition = function() {
		// override
		return "right";
	};

	this.getRestrictBorder = function() {
		return 0;
	};

	this.update();
};

// inherit
TextSelectionPopoverModel.prototype = Object.create(ViewportPopoverModel.prototype);
TextSelectionPopoverModel.prototype.constructor = TextSelectionPopoverModel;

/**
 * Update the popover
 */
TextSelectionPopoverModel.prototype.update = function() {
	_.extend(this.t.d, this._options);
	this.t.render();
};

/**
 * Set the list of areas of the highlight popover
 * @param {Array.<ShapeUtils.Rect>} The array of areas that overlap the individual text tags
 */
TextSelectionPopoverModel.prototype.setWorldTextFrames = function(worldTextFrames) {
	this._worldTextFrames = worldTextFrames;
	// Calculate the bounding rect which contains all these text frames
	this._boundingWorldFrame = this._calculateBoundingWorldFrame(worldTextFrames);
};

/**
 * Set the location of the popover
 * @param {ShapeUtils.Point} point for which the popover should be shown
 */
TextSelectionPopoverModel.prototype.setPopoverLocation = function(point) {
	this._popoverLocation = new ShapeUtils.Rect(point, new ShapeUtils.Size(0.001, 0.001));
};

/**
 * Set the area of the popover
 * @param {ShapeUtils.Rect} area for which the popover should be shown
 */
TextSelectionPopoverModel.prototype.setTextContent = function(content) {
	this._textContent = content;
};

TextSelectionPopoverModel.prototype.setPopoverOptions = function(options) {
	this._options = _.extend({ addTextAnnotation: false, addTextHighlightAnnotation: false, copyText: true, editText: false }, options);
	this.update();
};

/**
 * Get the bounding world frame, based on given the array of world text rects
 * @private
 * @param   {Array.<ShapeUtils.Rect>} worldTextFrames - the array of individual rects
 * @returns {ShapeUtils.Rect} the bounding world rect
 */
TextSelectionPopoverModel.prototype._calculateBoundingWorldFrame = function(worldTextFrames) {
	var boundingWorldFrame = null;
	if (worldTextFrames) {
		_.each(
			worldTextFrames,
			_.bind(function(textFrame) {
				if (!boundingWorldFrame) {
					boundingWorldFrame = textFrame;
				} else {
					boundingWorldFrame = boundingWorldFrame.calculateArea(textFrame);
				}
			}, this)
		);
	}
	return boundingWorldFrame;
};

export default TextSelectionPopoverModel;
