import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import LayoutConstants from "./LayoutConstants";
import LayoutReducer from "./LayoutReducer";
import { addItem, addLayoutModes, removeItem } from "./LayoutActions";
import { selectItem, selectLayoutModeByName } from "./LayoutSelect";
import LayoutItem from "@esko/web-viewer/layout/LayoutItem";

/**
 *
 * @exports LayoutModel
 * @memberof Layout
 */
var LayoutModel = function(f) {
	PrivateChannel.installTo(this, f, ["layoutModel"]);
	this.f = f;
	f.addReducer(LayoutReducer);
};

/**
 * Checks if the layoutItem with itemName is part of the layoutModel
 * @param {String} itemName
 * @returns {Boolean}
 */
LayoutModel.prototype.hasItem = function(itemName) {
	return this.f.select(selectItem, itemName) !== undefined;
};

/**
 * Add a widget to the layoutModel
 * @param {LayoutConstants.Widget} name - the name of the widget
 * @param {Function} widgetFactory - factory function to instantiate the widget (with 1 parameter: LayoutConstants.LayoutType)
 */
LayoutModel.prototype.addWidget = function(name, widgetFactory) {
	this.f.dispatch(addItem(name, LayoutConstants.LayoutItemType.WIDGET, widgetFactory));
	var newItem = new LayoutItem(name, LayoutConstants.LayoutItemType.WIDGET, widgetFactory);

	this._emit("itemAdded", newItem);
};

LayoutModel.prototype.removeItem = function(name) {
	this.f.dispatch(removeItem(name));
};

/**
 * Add a toolbarComponent to the layoutModel
 * @param {LayoutConstants.ToolbarComponent} name - the name the toolbarcomponent
 * @param {Function} toolbarComponentFactory - factory function to instantiate the toolbarcomponent (with 1 parameter: LayoutConstants.LayoutType)
 */
LayoutModel.prototype.addToolbarComponent = function(name, toolbarComponentFactory) {
	this.f.dispatch(addItem(name, LayoutConstants.LayoutItemType.TOOLBARCOMPONENT, toolbarComponentFactory));
	var newItem = new LayoutItem(name, LayoutConstants.LayoutItemType.TOOLBARCOMPONENT, toolbarComponentFactory);

	this._emit("itemAdded", newItem);
};

/**
 * Get the layoutType of the layoutItem
 * @param {String} itemName
 * @returns {LayoutConstants.LayoutItemType}
 */
LayoutModel.prototype.getLayoutItemType = function(itemName) {
	const item = this.f.select(selectItem, itemName);
	return item ? item.type : undefined;
};

/**
 * Get the layout item.
 * @param {String} itemName
 * @returns {Object}
 */
LayoutModel.prototype.getLayoutItem = function(itemName) {
	const item = this.f.select(selectItem, itemName);
	return item ? item.getItem() : undefined;
};

/**
 * Register a layoutMode to the model. All available mode for the loaded layout are saved in the model.
 * @param {LayoutMode} mode
 */
LayoutModel.prototype.registerLayoutMode = function(mode) {
	this.f.dispatch(addLayoutModes([mode]));
};

/**
 * Get the LayoutMode.
 * @param {LayoutConstants.LayoutMode} modeName
 * @returns {LayoutMode}
 */
LayoutModel.prototype.getLayoutMode = function(modeName) {
	return this.f.select(selectLayoutModeByName, modeName);
};

export default LayoutModel;
