import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
import ConversionConstants from "../utils/ConversionConstants";
import RBush from "rbush";
import rbushknn from "rbush-knn";
import { Character } from "./TextSelectionModel";
import { isRTL } from "./TextUtils";
const getFontMap = svg => {
	var style = svg.querySelector("style");
	var fontMap = {};
	if (style) {
		// SVG from NDL include font information in a style tag. for Chrome and FF we can use browser functionality to get the info, but IE gives us the wrong information.
		// we therefor parse the style tag ourselves to get the info.
		try {
			fontMap = _.object(
				style.textContent
					.trim()
					.split("\n")
					.map(x =>
						x
							.replace(/^\./, "")
							.replace(/;}$/, "")
							.split("{font-family:")
					)
			);
		} catch (e) {
			console.log("could not parse style tag");
			console.log(e);
			fontMap = {};
		}
	}
	return fontMap;
};

/** help function to get all font face data urls from within an svg */
const getFontDataURLs = (svg) => {
	const styleTags = svg.getElementsByTagName("style");
	const regex = new RegExp(/@font-face[ ]?{font-family:[^;]*;src:url\("([^"]+)[^}]+}/g)
	return _.flatten(Array.prototype.map.call(styleTags, style => {
		const content = style.textContent; 
		const fonts = [];
		var array;
		while((array = regex.exec(content)) != null) {
			fonts.push(array[1])
		}
		return fonts;
	}));
}

/**
 * To generate the characters used in TextSelectionModel from an SVG, we add the SVG to the DOM
 * and calculate the position of each <text> tag by using browser functionality
 */
var getCharactersFromSVG = function(svg) {
	if (svg) {
			var dummyDiv = document.createElement("div");
			dummyDiv.style.position = "fixed";
			dummyDiv.style.opacity = 0;
			dummyDiv.style.zIndex = -1;
			dummyDiv.style.top = 0;
			dummyDiv.style.left = 0;
			svg = svg.cloneNode(true);

			dummyDiv.appendChild(svg);
			document.body.appendChild(dummyDiv);

			return Promise.all(getFontDataURLs(svg).map(url => {
				// We have to wait until all data urls are loaded so that the document is rendered correctly.
				// There is no native way to listen to this so we create link preload elements and wait for those to load.
					return new Promise((resolve) => {
						const link = document.createElement("link");
						link.href = url;
						link.as = "font";
						link.rel = "preload";
						link.onload = resolve;
						link.onerror = resolve; // on error we'll try anyhow
						document.head.appendChild(link);
					});

				
			})).then(() => {
				
				var fontMap = getFontMap(svg);
				var rbushElems = [];
				/** for every character inside the text element we need to repeat the Y coordinate to make sure hebrew, arabic etc texts are rendered correctly
				 * This should actually be done by NDL but as it isn't done right now we do it here.
				 */
				_.each(document.querySelectorAll("text"), el => {
					const numberOfCharacters = el.getNumberOfChars();
					var ys = el.getAttribute("y") ? el.getAttribute("y").split(" ") : [];
					if (ys.length > 0) {
						for (let i = ys.length; i < numberOfCharacters; i++) {
							ys.push(ys[i - 1]);
						}
					}
					el.setAttribute("y", ys.join(" "));
				});

				try {
					rbushElems = calculateRbushElementsSVG(svg, fontMap);
				} catch (e) {
					console.error(e);
					console.error("rbush - could not insert character bounding boxes");
				}

				dummyDiv.parentElement.removeChild(dummyDiv);
				return rbushElems;
			});
	} else {
		return Promise.resolve([]);
	}
};

/**
 * @param {SVGElement} svgElement
 * @param {Object} fontMap
 */
var calculateRbushElementsSVG = function(svgElement, fontMap) {
	const rbush = new RBush(); // We use an rbush to  look for duplicates text tags (stroke and fill text tags)
	var rbushElems = [];
	_.forEach(svgElement.querySelectorAll("text"), (el, index) => {
		const numberOfCharacters = el.getNumberOfChars();
		var textContentTrimmed = el.textContent.trim();

		var { fontSize } = getComputedStyle(el); // default to  font size as computed by the browser

		// NDL uses a class to indicate which font is used. Check if that is available for this text tag and use that font family if so
		var clazz = el.getAttribute("class");
		let fontFamily = "";
		if (clazz && fontMap[clazz]) {
			fontFamily = fontMap[clazz];
		}

		// Check if there is a more specific font size than the one computed by the browser.
		var size = el.getAttribute("font-size")
		if (size) {
			fontSize = size;
		}
		fontSize = parseFloat(fontSize);

		const getChar = i => {
			// index i is the position of a character inside the text tag
			var extent = el.getExtentOfChar(i);
			var coords = convertCoords(extent, el, svgElement);
			return new Character(
				coords.origin.x,
				coords.origin.y,
				coords.origin.x + coords.size.width,
				coords.origin.y + coords.size.height,
				rbushElems.length,
				textContentTrimmed[i],
				index,
				fontFamily,
				fontSize
			);
		};

		let range;
		if (isRTL(textContentTrimmed)) {
			range = _.range(numberOfCharacters - 1, -1);
		} else {
			range = _.range(0, numberOfCharacters);
		}
		_.each(range, j => {
			let obj = getChar(j);
			if (obj.character) {
				// We want to check if there is no duplicate character at the same position. We do this by using rbushknn with a 'small' distance.
				const d = obj.rect.topright().getDistance(obj.rect.bottomleft()) / 100; // small distance is 1% of the diagonal of the character
				var [elem] = rbushknn(rbush, obj.rect.origin.x, obj.rect.origin.y, 1, undefined, d);

				if (!elem || elem.character !== obj.character) {
					// No element was found or the close element has a different character. We will use it for text selection.
					const el = {
						minX: obj.minX,
						maxX: obj.minX,
						minY: obj.minY,
						maxY: obj.minY,
						character: obj.character
					};
					rbush.insert(el);
					rbushElems.push(obj);
				}
			}
		});
	});
	return rbushElems;
};

// This funky function calculates the locadtion of a character inside the svg
var convertCoords = function(extent, element, svg) {
	var matrix = element.getScreenCTM();

	var pt = svg.createSVGPoint();
	pt.x = extent.x;
	pt.y = extent.y;
	pt = pt.matrixTransform(matrix);

	var pt2 = svg.createSVGPoint();
	pt2.x = extent.x + extent.width;
	pt2.y = extent.y + extent.height;
	pt2 = pt2.matrixTransform(matrix);

	var r = ConversionConstants.PX_TO_MM_RATIO;

	var scenePoint = new ShapeUtils.Point(pt.x * r, pt.y * r);
	var sceneSize = new ShapeUtils.Size((pt2.x - pt.x) * r, (pt2.y - pt.y) * r);

	var rect = new ShapeUtils.Rect(scenePoint, sceneSize);
	rect.normalize();
	return rect;
};

/**
 * This  can be mixed into a document page to ensure a page can ues text selection functionalty
 */
export default {
	/**
	 * generates text information to be used in the TextSelectionModel from an SVG element.
	 * @param {AbstractDocumentPage} page - the page to which the text data will be added
	 * @param {SVGElement} svg - element that is used to generate the text information
	 */
	fromSVG: function(page, svg) {
		getCharactersFromSVG(svg).then(chars => {
			page.set("textData", chars);
		});
	},
	/**
	 * NDL/TextScraper currently outputs an SVG so we just use the same algorithm as for SVG view files
	 */
	fromNDL: function(page, svg) {
		getCharactersFromSVG(svg).then(chars => page.set("textData", chars));
	}
};
